#!/usr/bin/env python

import collections
import iminuit
import inspect
import io
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import re
import scipy.odr
import scipy.signal
import sys


class Histogram(object):

    @classmethod
    def _asarray(cls, a):
        a = np.asarray(a, dtype=np.float64)
        a.setflags(write=False)
        return a

    def __init__(self, *, edges, values, variances=None, axislabel=None):
        self.edges = edges = self._asarray(edges)
        self.values = values = self._asarray(values)
        self.variances = variances = (values if variances is None else self._asarray(variances))
        self.axislabel = axislabel


    @property
    def centers(self):
        return (self.edges[:-1] + self.edges[1:])/2.
    
    @property
    def widths(self):
        return (self.edges[1:] - self.edges[:-1])
    
    @property
    def nbins(self):
        return len(self.values)

    def get_fit_data(self):
        return {'x': self.centers, 'y': self.values, 'ex': self.widths / 2., 'ey': self.errors}

    @property
    def errors(self):
        return np.sqrt(self.variances)

    def replace_axis(self, edges, axislabel=None):
        return Histogram(edges=edges, values=self.values, variances=self.variances, axislabel=axislabel)

    def rebin(self, *, start=None, stop=None, ngroup=None):
        edges = self.edges
        variances = self.variances
        values = self.values
        if not (start is None and stop is None):
            i = slice(start, stop)
            edges = np.concatenate([self.edges[:-1][i], self.edges[1:][i][-1:]])
            values = values[i]
            variances = variances[i]
        if ngroup is not None:
            m = len(values)
            i = np.repeat(np.arange(m//ngroup), ngroup)
            if len(i) != m:
                raise ValueError("n (%r) is not a divisor of the bin count (%r)" % (ngroup, m))
            edges = edges[::ngroup]
            values = np.bincount(i, weights=values)
            variances = np.bincount(i, weights=variances)
        return Histogram(edges=edges, values=values, variances=variances, axislabel=self.axislabel)


class FitResult(collections.namedtuple('FitResultTuple', 'minuit names values covariance fcn fval data method is_fixed curve')):

    def _repr_html_(self):
        self.minuit._repr_html_()
        #return '\n'.join([
        #    self.minuit.get_fmin()._repr_html_(),
        #    self.minuit.get_param_states()._repr_html_(),
        #    self.minuit.matrix(correlation=True, skip_fixed=True)._repr_html_()])

    def to_params(self):
        params = {}
        for i, name in enumerate(self.names):
            params[name] = self.values[i]
            params['error_' + name] = np.sqrt(self.covariance[i,i])
            params['fix_' + name] = self.is_fixed[i]
        return params


def load_spectrum(path):
    state = 0
    start, stop = None, None
    values = None
    with open(path, 'rb') as f:
        for line in f:
            m = re.match(br'^\$(\w+):\s*$', line)
            if m:
                if state == 1:
                    state = 2
                if m.group(1) == b'DATA':
                    assert state == 0
                    state = 1
            elif state == 1:
                row = line.split()
                if values is None:
                    start, stop = row
                    start = int(start)
                    stop = int(stop)
                    values = []
                else:
                    value, = row
                    values.append(value)
    assert state == 2
    assert len(values) == stop - start + 1
    return Histogram(edges=np.arange(len(values)+1)-0.5, values=values, axislabel='Channel Number')


def find_peaks(a):
    peaks = scipy.signal.find_peaks_cwt(a, [10.])
    xs = []
    for peak in peaks:
        i, j = max(peak - 4, 0), min(peak + 5, len(a))
        xmax = i + np.argmax(a[i:j])
        xs.append(xmax)
    return np.asarray(xs)


def weighted_least_squares(x, y, ey, curve):
    if not (ey > 0.).all():
        raise ValueError("errors must be greater than zero")
    def fcn(*args):
        expectation = curve(*args)(x)
        return np.square((expectation - y) / ey).sum()
    return fcn


def chi_squared(x, y, curve):
    invalid = np.finfo(np.float64).max
    def fcn(*args):
        expectation = curve(*args)(x)
        if not (expectation > 0.).all():
            return invalid
        return (np.square(expectation - y) / expectation).sum()
    return fcn


def fit(*, method, curve, data, params):
    if hasattr(data, 'get_fit_data'):
        data = data.get_fit_data()
    method_params = list(inspect.signature(method).parameters)
    data = {k: np.asarray(v, dtype=np.float64) for k, v in data.items() if k in method_params}
    fcn = method(curve=curve, **data)
    iminuit.Minuit.errdef=1.0
    minuit = iminuit.Minuit(fcn, name=inspect.signature(curve).parameters.keys(), **params)
    minuit.migrad()
    #if not minuit.migrad_ok():
    #    raise RuntimeError("fit failed")
    minuit.hesse()
    names = tuple(p for p in inspect.signature(curve).parameters.keys())
    values = np.asarray(minuit.values, dtype=np.float64)
    values.setflags(write=False)
    covariance = minuit.covariance
    covariance.setflags(write=False)
    return FitResult(minuit=minuit,
                     method=method,
                     names=names,
                     values=values,
                     covariance=covariance,
                     is_fixed=np.asarray(minuit.fixed, dtype=bool),
                     curve=curve,
                     fcn=fcn,
                     data=data,
                     fval=minuit.fval)


def odr(*, curve, data, ignore_xerrors=False):
    names = tuple(inspect.signature(curve).parameters)
    data = {k: np.asarray(v, dtype=np.float64) for k, v in data.items()}
    myfit = scipy.odr.ODR(scipy.odr.RealData(**data), lambda B, x: curve(*B)(x))
    myfit.set_job(var_calc=0)
    if ignore_xerrors:
        myfit.set_job(fit_type=2)
    myoutput = myfit.run()
    if myoutput.info != 1:
        raise RuntimeError("fit failed")
    myoutput.beta.setflag(write=False)
    myoutput.cov_beta.setflag(write=False)
    return FitResult(names=names, values=myoutput.beta, covariance=myoutput.cov_beta, curve=curve)
